// Package squashfs provides support for reading and creating squashfs filesystems
// references:
//    https://www.kernel.org/doc/Documentation/filesystems/squashfs.txt
//    https://dr-emann.github.io/squashfs/
//    https://elinux.org/images/3/32/Squashfs-elce.pdf
package squashfs
