// Copyright 2018 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package internal

// EnableTrace indicates whether stack information should be recorded in errors.
var EnableTrace = true
