// Package fat32 provides utilities to interact with, manipulate and create a FAT32 filesystem on a block device or
// a disk image.
//
// references:
//   https://en.wikipedia.org/wiki/Design_of_the_FAT_file_system
//   https://www.cs.fsu.edu/~cop4610t/assignments/project3/spec/fatspec.pdf
//   https://wiki.osdev.org/FAT
//
package fat32
