// Package iso9660 provides utilities to interact with, manipulate and create an iso9660 filesystem on a block device or
// a disk image.
//
// Reference documentation
//  ISO9660 https://wiki.osdev.org/ISO_9660
//  ISO9660 / ECMA-119 http://www.ecma-international.org/publications/files/ECMA-ST/Ecma-119.pdf
//  System Use Sharing Protocol http://cdrtools.sourceforge.net/private/RRIP/susp.ps
//  Rock Ridge http://cdrtools.sourceforge.net/private/RRIP/rrip.ps
//  El Torito https://wiki.osdev.org/El-Torito
package iso9660
